#include <openssl/conf.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <sys/socket.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <tlssep.h>
#include <tlssep-config.h>
#include <tlssep-private.h>
#include <tlssep-verify.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define DESC_SIZE          1024
#define SELECT_LIMIT       1024 /* Max number of F.D.s in fd_set? */
#define OP_SOCKET          STDERR_FILENO + 1
#define FD_TRANSFER_SOCKET STDERR_FILENO + 2

typedef struct desc_t {
	int  tlsFd;
	SSL *ssl;
	int  notificationSocket;
	bool blocking_io;
	bool notified;
} desc_t;

fd_set      rfds;
const char *DEFAULT_CONFIG_PATH = SYSCONFDIR "/tlssep-decorator.cfg";
bool        running             =  true;
desc_t      desc[DESC_SIZE];

static void
_reset_desc_entry(unsigned int i)
{
	desc[i].tlsFd                     = -1;
	desc[i].ssl                       =  NULL;
	desc[i].notificationSocket        = -1;
	desc[i].blocking_io               =  true;
	desc[i].notified                  =  false;
}

static void
_init_svc(tlssep_args_t *args)
{
	static tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };
	
	SSL_library_init();
	SSL_load_error_strings();
	OPENSSL_config(NULL); // FIXME: Default configuration settings.

	int ret = write(OP_SOCKET, &retval, sizeof(tlssep_ret_t));
        if (-1 == ret) {
		perror("Failed to write init response");
		exit(EXIT_FAILURE);
        }
}

static int
_verify_null_callback(int preverify_ok, X509_STORE_CTX *ctx)
{
	/* NOTE: we handle verification later, after the connection.
	 * It would be nice to handle it here, but sadly this OpenSSL
	 * callback does not take as an argument an arbitrary state.
	 * This makes it difficult to verify using a novel technique.
	 */
	return true;
}

static tlssep_status_t
_tls_init(const char *cert_path, const char *key_path, const SSL_METHOD *(method_fn) (void), int tlsFd, SSL_CTX **ctx, SSL **ssl)
{
	assert(TLSv1_2_client_method == method_fn
	    || TLSv1_2_server_method == method_fn);
	assert(NULL != ctx);
	assert(NULL != ssl);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	
	const SSL_METHOD *meth = method_fn();
	if(NULL == meth) {
		status = TLSSEP_STATUS_ERROR_TLS;
		goto done;
	}

	if (NULL == *ctx) {
		*ctx = SSL_CTX_new(meth);
		if(NULL == *ctx) {
			status = TLSSEP_STATUS_ERROR_TLS;
			goto done;
		}

		bool require_client_certificate = false;
		if (true == tlssep_config_lookup_require_client_certificate(&require_client_certificate)
		 && true == require_client_certificate
		 && TLSv1_2_server_method == method_fn) {
			SSL_CTX_set_verify(*ctx, SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT, _verify_null_callback);
		}

		if(cert_path && SSL_CTX_use_certificate_file(*ctx, cert_path, SSL_FILETYPE_PEM) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
			goto done;
		}

		if(key_path && SSL_CTX_use_PrivateKey_file(*ctx, key_path, SSL_FILETYPE_PEM) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
			goto done;
		}

		if(key_path && SSL_CTX_check_private_key(*ctx) != 1) {
			status = TLSSEP_STATUS_ERROR_TLS;
			goto done;
		}

		if(SSL_CTX_set_default_verify_paths (*ctx) != 1) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CA_PATH;
			goto done;
		}
	}

	*ssl = SSL_new(*ctx);
	if(NULL == *ssl) {
		status = TLSSEP_STATUS_ERROR_TLS;
		goto done;
	}

	int sslFD = SSL_set_fd(*ssl, tlsFd);
	if(0 == sslFD) {
		status = TLSSEP_STATUS_ERROR_TLS;
		goto done;
	}
	
done:
	if(TLSSEP_STATUS_OK != status) {
		if (NULL != *ctx) {
			SSL_CTX_free(*ctx);
			*ctx = NULL;
		}

		if (NULL != *ssl) {
			SSL_free(*ssl);
			*ssl = NULL;
		}
        } 
	
	return status;
}

// http://blog.varunajayasiri.com/passing-file-descriptors-between-processes-using-sendmsg-and-recvmsg:
static int
recv_file_descriptor(int socket)
{
	int fnval = -1;
	struct msghdr message;
	struct iovec iov[1];
	struct cmsghdr *control_message = NULL;
	char ctrl_buf[CMSG_SPACE(sizeof(int))];
	char data[1];
	int res;

	memset(&message, 0, sizeof(struct msghdr));
	memset(ctrl_buf, 0, CMSG_SPACE(sizeof(int)));

	/* For the dummy data */
	iov[0].iov_base = data;
	iov[0].iov_len  = sizeof(data);

	message.msg_name       = NULL;
	message.msg_namelen    = 0;
	message.msg_control    = ctrl_buf;
	message.msg_controllen = CMSG_SPACE(sizeof(int));
	message.msg_iov        = iov;
	message.msg_iovlen     = 1;

	if((res = recvmsg(socket, &message, 0)) <= 0) {
		return res;
	}

	/* Iterate through header to find if there is a file descriptor. */
	for(control_message = CMSG_FIRSTHDR(&message); control_message != NULL; control_message = CMSG_NXTHDR(&message, control_message)) {
		if((control_message->cmsg_level == SOL_SOCKET) && (control_message->cmsg_type == SCM_RIGHTS)) {
			memcpy(&fnval, CMSG_DATA(control_message), sizeof(fnval));
			goto done;
		}
	}

done:
	return fnval;
}

static void
_connect_or_accept_svc(tlssep_args_t *args, int (*SSL_op) (SSL *ssl))
{
	assert(SSL_op == SSL_connect || SSL_op == SSL_accept);

	int ret;
	static SSL_CTX *ctx =  NULL;
	int    sslCon       = -1;

	static tlssep_ret_t fnval = {
		.status = TLSSEP_STATUS_OK,
		.id = -1,
	};

	for (fnval.id = 0; fnval.id < DESC_SIZE; fnval.id++) {
		if (desc[fnval.id].tlsFd == -1) {
			break;
		}
	}

	if (DESC_SIZE == fnval.id) {
		fnval.status = TLSSEP_STATUS_ERROR_DESCRIPTORS_EXHAUSTED;
		goto done;
	}

	assert(NULL == desc[fnval.id].ssl);

	desc[fnval.id].tlsFd = recv_file_descriptor(FD_TRANSFER_SOCKET);
	if (desc[fnval.id].tlsFd <= 0
	 || desc[fnval.id].tlsFd >= SELECT_LIMIT) {
		fnval.status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	desc[fnval.id].notificationSocket = recv_file_descriptor(FD_TRANSFER_SOCKET);
	if (desc[fnval.id].notificationSocket <= 0) {
		fnval.status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	if (SSL_op == SSL_connect) {
		const char *certpath = NULL;
		const char *privkeypath = NULL;
		bool provide_client_certificate = false;
		if (true == tlssep_config_lookup_provide_client_certificate(&provide_client_certificate)
		 && true == provide_client_certificate) {
			if (false == tlssep_config_lookup_certificate_path(&certpath)) {
				fnval.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
				goto done;
			}
			if (false == tlssep_config_lookup_privkey_path(&privkeypath)) {
				fnval.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
				goto done;
			}
		}
		fnval.status = _tls_init(certpath, privkeypath, TLSv1_2_client_method, desc[fnval.id].tlsFd, &ctx, &desc[fnval.id].ssl);
	} else if (SSL_op == SSL_accept) {
		const char *certpath = NULL;
		const char *privkeypath = NULL;
		if (false == tlssep_config_lookup_certificate_path(&certpath)) {
			fnval.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_CERT_PATH;
			goto done;
		}
		if (false == tlssep_config_lookup_privkey_path(&privkeypath)) {
			fnval.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_PRIVKEY_PATH;
			goto done;
		}
		fnval.status = _tls_init(certpath, privkeypath, TLSv1_2_server_method, desc[fnval.id].tlsFd, &ctx, &desc[fnval.id].ssl);
	} else {
		assert(false);
	}

	if(TLSSEP_STATUS_OK != fnval.status) {
		goto done;
	}

	sslCon = SSL_op(desc[fnval.id].ssl);
	if (1 != sslCon) {
		fnval.status = SSL_op == SSL_connect
		             ? TLSSEP_STATUS_ERROR_TLS_CONNECT
		             : TLSSEP_STATUS_ERROR_TLS_ACCEPT;
		goto done;
	}

	int engine_mask = 0;
	if(false == tlssep_config_lookup_verification_engines(&engine_mask)) {
		fnval.status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_VERIFICATION_ENGINE;
		goto done;
	}

	int enough;
	if (false == tlssep_config_lookup_enough(&enough)) {
		/* Fall back on requiring all configured engines. */
		int i;
		enough = 0;
		for (i = 0; i < sizeof(engine_mask) * 8; i++) {
			if (engine_mask & 1 << i) {
				enough++;
			}
		}
	}

	assert(sizeof(args->buf) > TLSSEP_MAX_COMMON_NAME_SIZE);
	fnval.status = tlssep_certificate_ok(desc[fnval.id].ssl, engine_mask, enough, args->buf, true == args->size ? fnval.buf : NULL);
	if(TLSSEP_STATUS_OK != fnval.status) {
		goto done;
	}

	fnval.size = strlen(fnval.buf);

	FD_SET(desc[fnval.id].tlsFd, &rfds);

done:
	ret = write(OP_SOCKET, &fnval, sizeof(fnval));
	if (-1 == ret) {
		perror("Failed to write connect/accept response");
		exit(EXIT_FAILURE);
	}

	if(TLSSEP_STATUS_OK != fnval.status) {
		if (-1 == desc[fnval.id].tlsFd) {
			close(desc[fnval.id].tlsFd);
		}

		if (-1 == desc[fnval.id].notificationSocket) {
			close(desc[fnval.id].notificationSocket);
		}

		if (NULL != ctx) {
			SSL_CTX_free(ctx);
			ctx = NULL;
		}

		if (fnval.id < DESC_SIZE && NULL != desc[fnval.id].ssl) {
			SSL_free(desc[fnval.id].ssl);
		}

		_reset_desc_entry(fnval.id);
        }
}

static tlssep_status_t
_write_some(SSL *ssl, char *buf, int size)
{
	tlssep_status_t fnval = TLSSEP_STATUS_OK;

	while (true) {
		int bytes_written = SSL_write(ssl, buf, size);
		if (bytes_written == size) {
			break;
		} else if(-1 == bytes_written) {
			if (SSL_ERROR_WANT_WRITE == SSL_get_error(ssl, bytes_written)) {
				continue; /* Try again. */
			} else {
				fnval = TLSSEP_STATUS_ERROR_TLS_WRITE;
				goto done;
			}
		}
	}

done:
	return fnval;
}

static void
_write_svc(tlssep_args_t *args)
{
	int ret;
	static tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };
	char *buf = NULL;

	assert(args->size > 0);

	int remaining = args->size;
	int to_write = args->size > BUFSIZ ? BUFSIZ : args->size;

	retval.status = _write_some(desc[args->id].ssl, args->buf, to_write);
	if (TLSSEP_STATUS_OK != retval.status) {
		goto done;
	}

	remaining -= to_write;
	int bytes_written = to_write;

	if (0 != remaining) {
		/* args structure did not contain all data to write. */
		buf = malloc(remaining);
		if (NULL == buf) {
			retval.status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}

		do {
			int to_write = read(OP_SOCKET, buf, remaining);
			if (-1 == to_write) {
				perror("Failed to read write buffer");
				exit(EXIT_FAILURE);
			}

			retval.status = _write_some(desc[args->id].ssl, buf, to_write);
			if (TLSSEP_STATUS_OK != retval.status) {
				goto done;
			}

			remaining -= to_write;
			bytes_written += to_write;
		} while (0 != remaining);
	}

	retval.size   = args->size;
	retval.status = TLSSEP_STATUS_OK;

done:
        ret = write(OP_SOCKET, &retval, sizeof(retval));
        if (sizeof(retval) != ret) {
		perror("Failed to write write return value");
		exit(EXIT_FAILURE);
        }

	if (NULL != buf) {
		free(buf);
	}
}

static void
_read_or_peek_svc(tlssep_args_t *args, int (*SSL_op) (SSL *ssl, void *buf, int num))
{
	assert(SSL_read == SSL_op || SSL_peek == SSL_op);

	int readLen = 0;
	char *buf = NULL;
	static tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	assert(args->size > 0);

	desc[args->id].notified = false;

	if (args->size <= BUFSIZ) {
		buf = retval.buf;
	} else {
		buf = malloc(args->size);
		if (NULL == buf) {
			retval.status = TLSSEP_STATUS_ERROR_ALLOCATION_FAILED;
			goto done;
		}
	}

	readLen = SSL_op(desc[args->id].ssl, buf, args->size);
	if(readLen < 0) {
		if (desc[args->id].blocking_io == false && errno == EAGAIN) {
			retval.status = TLSSEP_STATUS_AGAIN;
			goto done;
		} else {
			retval.status = TLSSEP_STATUS_ERROR_TLS_READ;
			goto done;
		}
	}

done:
	retval.size = readLen > 0 ? readLen : 0;

	if (retval.size > 0 && retval.size <= BUFSIZ && buf != retval.buf) {
		/* Read did not end up using more than BUFSIZ bytes. */
		memcpy(retval.buf, buf, retval.size);
	}

	int ret = write(OP_SOCKET, &retval, sizeof(retval));
	if (sizeof(retval) != ret) {
		perror("Failed to write read/peek return value");
		exit(EXIT_FAILURE);
	}

	if (retval.size > BUFSIZ) {
		ret = write(OP_SOCKET, buf, retval.size);
		if (retval.size != ret) {
			perror("Failed to write read/peek return buffer");
			exit(EXIT_FAILURE);
		}
	}

	if (buf != retval.buf) {
		free(buf);
	}
}

static void
_setnonblock_svc(tlssep_args_t *args)
{
	tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	int ret = fcntl(desc[args->id].tlsFd, F_SETFL, O_NONBLOCK);

	if (ret != -1) {
		desc[args->id].blocking_io = false;
		retval.status = TLSSEP_STATUS_OK;
	} else {
		retval.status = TLSSEP_STATUS_ERROR;
	}

	ret = write(OP_SOCKET, &retval, sizeof(tlssep_ret_t));
        if (-1 == ret) {
		perror("Failed to write setnonblock return values");
		exit(EXIT_FAILURE);
        }

	return;
}

static void
_close_svc(tlssep_args_t *args)
{
	int ret;
	tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	assert(desc[args->id].ssl != NULL);

	FD_CLR(desc[args->id].tlsFd, &rfds);

	int shutdown = SSL_shutdown(desc[args->id].ssl);
	if (0 == shutdown) {
		SSL_shutdown(desc[args->id].ssl);
	} else if(shutdown < 0) {
		retval.status = TLSSEP_STATUS_ERROR;
		goto done;
	}

	SSL_free(desc[args->id].ssl);

	int result = close(desc[args->id].tlsFd);
	if(0 != result) {
		retval.status = TLSSEP_STATUS_ERROR;
		goto done;
	}

	if (-1 != desc[args->id].notificationSocket) {
		result = close(desc[args->id].notificationSocket);
		if (0 != result) {
			retval.status = TLSSEP_STATUS_ERROR;
			goto done;
		}
	}

done:
	ret = write(OP_SOCKET, &retval, sizeof(tlssep_ret_t));
        if (sizeof(tlssep_ret_t) != ret) {
		perror("Failed to write close return values");
		exit(EXIT_FAILURE);
        }

	_reset_desc_entry(args->id);

}

static void
_terminate_svc(tlssep_args_t *args)
{
	running = false;

	tlssep_config_free();

	if (-1 != OP_SOCKET) {
		int result = close(OP_SOCKET);
		if (0 != result) {
			fprintf(stderr, "%s\n", "Error closing file descriptor transfer socket");
		}
	}

	if (-1 != FD_TRANSFER_SOCKET) {
		int result = close(FD_TRANSFER_SOCKET);
		if (0 != result) {
			fprintf(stderr, "%s\n", "Error closing file descriptor transfer socket");
		}
	}

	exit(EXIT_SUCCESS);
}

static void
_notify_fd_set(fd_set rfds, desc_t *desc)
{
	int i;
	for (i = 0; i < DESC_SIZE; i++) {
		if (desc[i].tlsFd != -1 && FD_ISSET(desc[i].tlsFd, &rfds)) {
			if (desc[i].tlsFd != -1 && false == desc[i].notified) {
				assert(-1 != desc[i].notificationSocket);
				if (-1 == write(desc[i].notificationSocket, "x", 1)) {
					perror("Error writing to notification socket");
					exit(EXIT_FAILURE);
				}
				desc[i].notified = true; /* Do not notify again until after a read_op. */
			}
		}
	}
}

static void
_handle_op(int *left, fd_set rfds)
{
	if (FD_ISSET(OP_SOCKET, &rfds)) {
		static tlssep_args_t args;
		int ret = read(OP_SOCKET, &args, sizeof(args));
                if (sizeof(args) != ret) {
			perror("Failed to read operation (did application forget to terminate decorator?)");
			exit(EXIT_FAILURE);
                }

		*left = *left - 1;
	
		switch(args.op) {
		case TLSSEP_OP_INIT:
			_init_svc(&args);
			break;
		case TLSSEP_OP_CONNECT:
			_connect_or_accept_svc(&args, SSL_connect);
			break;
		case TLSSEP_OP_ACCEPT:
			_connect_or_accept_svc(&args, SSL_accept);
			break;
		case TLSSEP_OP_WRITE:
			_write_svc(&args);
			break;
		case TLSSEP_OP_READ:
			_read_or_peek_svc(&args, SSL_read);
			break;
		case TLSSEP_OP_PEEK:
			_read_or_peek_svc(&args, SSL_peek);
			break;
		case TLSSEP_OP_SETNONBLOCK:
			_setnonblock_svc(&args);
			break;
		case TLSSEP_OP_CLOSE:
			_close_svc(&args);
			break;
		case TLSSEP_OP_TERMINATE:
			_terminate_svc(&args);
			break;
		default:
			fprintf(stderr, "Received bad operation\n");
			exit(EXIT_FAILURE);
		}
	}
}

int
main (int argc, char *argv[])
{
	if (argc != 2) {
		fprintf(stderr, "Usage: %s APPLICATION_PATH\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	const char *applicationPath = argv[1];

	if (-1 == fcntl(OP_SOCKET, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive op socket from parent.\n");
		exit(EXIT_FAILURE);
	}

	if (-1 == fcntl(FD_TRANSFER_SOCKET, F_GETFD)) {
		fprintf(stderr, "%s", "Did not receive file descriptor transfer socket from parent.\n");
		exit(EXIT_FAILURE);
	}

	int i;
	for (i = 0; i < DESC_SIZE; i++) {
		_reset_desc_entry(i);
	}

	/* TODO: Hmm---is it safe to allow the application to set the config. path? */
	char *env_config_path = getenv("LIBTLSSEP_CONFIG_PATH");
	if (false == tlssep_config_init(NULL != env_config_path ? env_config_path : DEFAULT_CONFIG_PATH, applicationPath)) {
		fprintf(stderr, "Could not read %s\n", NULL != env_config_path ? env_config_path : DEFAULT_CONFIG_PATH);
                exit(EXIT_FAILURE);
	}

	/* Select on:
	 *	(1) RPC fd or RPC fds
	 *	(2) Network socket
	 * Handle RPC and, when network socket ready to read, write to notification 
	 * socket.
	 *
	 * Application can select on notification socket and RPC on RPC socket when
	 * select returns. Applications: DON'T FORGET TO READ BYTE from notification
	 * socket too!
	 */
	while (true == running) {
		int maxFd = OP_SOCKET;	

		FD_ZERO(&rfds);
		FD_SET(OP_SOCKET, &rfds);

		for (i = 0; i < DESC_SIZE; i++) {
			/* NOTE: do not select on TLS F.D. if application has already been notified. */
			if (desc[i].tlsFd != -1 && false == desc[i].notified) {
				FD_SET(desc[i].tlsFd, &rfds);
				maxFd = desc[i].tlsFd > maxFd ? desc[i].tlsFd : maxFd;
			}
		}

		int count = select(maxFd + 1, &rfds, NULL, NULL, NULL);

		switch (count) {
		case -1:
			perror("Error calling select");
			// exit(EXIT_FAILURE);
		case 0:
			break;
		default:
			_handle_op(&count, rfds);
			_notify_fd_set(rfds, desc);
			break;
		}
	}

        exit(EXIT_SUCCESS);
}
