#include <sys/signalfd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tlssep.h>
#include <tlssep-private.h>
#include <unistd.h>
#include <wait.h>

#define PROGRAM_NAME "tlssep-decorator"

typedef struct sync_handler_t {
	bool     installed;
	sigset_t mask_old;
	int      sfd;
} sync_handler_t;

static void
_release_context(tlssep_context_t *context)
{
	if (-1 != context->opSocket) {
		close(context->opSocket);
	}

	if (-1 != context->fdTransferSocket) {
		close(context->fdTransferSocket);
	}

	context->opSocket             = -1;
	context->fdTransferSocket     = -1;
	context->decoratorPid         = -1;
}

static tlssep_status_t
_tear_down_decorator(tlssep_context_t *context)
{
	assert(NULL != context);

	tlssep_status_t status = TLSSEP_STATUS_OK;
	int child_status;
	int wait_status;

	if (-1 != context->opSocket) {
		tlssep_args_t args = { .op = TLSSEP_OP_TERMINATE };
		int ret = write(context->opSocket, &args, sizeof(args));
		if (sizeof(args) != ret) {
			status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		}
	}

	if (-1 != context->decoratorPid) {
		wait_status = waitpid(context->decoratorPid, &child_status, 0);

		if (wait_status != context->decoratorPid || child_status != 0) {
			status = TLSSEP_STATUS_ERROR_DECORATOR_BAD_EXIT;
		}
	}

	_release_context(context);

	return status;
}

static int
_shiftFd(int fd, int offset)
{
	int newFd = -1;

	if (fd != STDERR_FILENO + offset) {
		newFd = dup(fd);
		if (newFd < 0) {
			goto done;
		}

		close(fd);
	} else {
		newFd = fd;
	}

done:
	return newFd;
}

static int
_closeFdsExcept(int skip1, int skip2)
{
	int fnval = false;
	
	int i;
	/* Close all file descriptors except stdin, out, err, and two given sockets. */
	for (i = STDERR_FILENO + 1; i < getdtablesize(); i++) {
		if (i == skip1 || i == skip2) {
			continue;
		}

		if (-1 == close(i) && errno != EBADF) {
			/* Some strange problem; file descriptor
			 * valid (! EBADF), but close failed.
			 */
			goto done;
		}
	}

	skip1 = _shiftFd(skip1, 1);
	if (-1 == skip1) {
		goto done;
	}

	skip2 = _shiftFd(skip2, 2);
	if (-1 == skip2) {
		goto done;
	}

	assert(STDERR_FILENO + 1 == skip1);
	assert(STDERR_FILENO + 2 == skip2);

	fnval = true;

done:
	return fnval;
}

static const char *
_get_application_path(void)
{
	char *fnval = NULL;
	static char path[PATH_MAX];

	ssize_t size = readlink("/proc/self/exe", path, sizeof(path) - 1);
	if (-1 == size) {
		goto done;
	}

	path[size] = 0x00;
	fnval = path;

done:
	return fnval;
}

static tlssep_status_t
_fork_decorator(sync_handler_t *handler, tlssep_context_t *context, int childOpSocket, int childFdTransferSocket)
{
	tlssep_status_t status = TLSSEP_STATUS_OK;
	struct stat stat_buf;
	const char *application_path = NULL;

	if (-1 == stat(LIBEXECDIR "/" PROGRAM_NAME, &stat_buf)) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING;
		goto done;
	}

	if (0 == S_ISREG(stat_buf.st_mode) && 0 == (stat_buf.st_mode & 0111)) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_MISSING;
		goto done;
	}

	application_path = _get_application_path();
	if (NULL == application_path) {
		status = TLSSEP_STATUS_ERROR;
		goto done;
	}

	context->decoratorPid = fork();
	if (-1 == context->decoratorPid) {
		status = TLSSEP_STATUS_ERROR_DECORATOR_FORK;
		goto done;
	} else if (context->decoratorPid == 0) {
		/* Child. */
		if (false == _closeFdsExcept(childOpSocket, childFdTransferSocket)) {
			fprintf(stderr, "%s\n", "Error closing existing file descriptors");
			exit(EXIT_FAILURE);
		};

		execl(LIBEXECDIR "/" PROGRAM_NAME, PROGRAM_NAME, application_path, NULL);
		fprintf(stderr, "Error executing %s\n", LIBEXECDIR "/" PROGRAM_NAME);
		exit(EXIT_FAILURE);
	}

done:
	return status;
}

tlssep_status_t
tlssep_init (tlssep_context_t *context)
{
	assert(NULL != context);

	int sockpair[2];
	int childOpSocket = -1;
	int childFdTransferSocket = -1;
	tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	context->opSocket         = -1;
	context->fdTransferSocket = -1;
	context->decoratorPid     = -1;

	if (-1 == socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair)) {
		retval.status = TLSSEP_STATUS_ERROR_SOCKET_FAILED;
		goto done;
	}

	context->opSocket = sockpair[0];
	childOpSocket     = sockpair[1];

	if (-1 == socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair)) {
		retval.status = TLSSEP_STATUS_ERROR_SOCKET_FAILED;
		goto done;
	}

	context->fdTransferSocket = sockpair[0];
	childFdTransferSocket     = sockpair[1];

	sync_handler_t sync_handler;
	retval.status = _fork_decorator(&sync_handler, context, childOpSocket, childFdTransferSocket);
	if (TLSSEP_STATUS_OK != retval.status) {
		goto done;
	}

	static tlssep_args_t args = { .op = TLSSEP_OP_INIT };
	int ret = write(context->opSocket, &args, sizeof(args));
	if (sizeof(args) != ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	ret = read(context->opSocket, &retval, sizeof(tlssep_ret_t));
	if (sizeof(tlssep_ret_t) != ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

done:
	if (-1 != childOpSocket) {
		close(childOpSocket);
	}

	if (-1 != childFdTransferSocket) {
		close(childFdTransferSocket);
	}

	if (TLSSEP_STATUS_OK != retval.status) {
		if (-1 != context->decoratorPid) {
			kill(context->decoratorPid, SIGKILL); /* Make sure child is dead. */
		}
		_release_context(context);
	}

	return retval.status;
}

// http://blog.varunajayasiri.com/passing-file-descriptors-between-processes-using-sendmsg-and-recvmsg:
static int
sendFileDescriptor(int socket, int fd_to_send)
{
	struct msghdr message;
	struct iovec iov[1];
	struct cmsghdr *control_message = NULL;
	char ctrl_buf[CMSG_SPACE(sizeof(int))];
	char data[1];

	memset(&message, 0, sizeof(struct msghdr));
	memset(ctrl_buf, 0, CMSG_SPACE(sizeof(int)));

	/* We are passing at least one byte of data so that recvmsg() will not return 0 */
	data[0]         = 'x';
	iov[0].iov_base = data;
	iov[0].iov_len  = sizeof(data);

	message.msg_name       = NULL;
	message.msg_namelen    = 0;
	message.msg_iov        = iov;
	message.msg_iovlen     = 1;
	message.msg_controllen = CMSG_SPACE(sizeof(int));
	message.msg_control    = ctrl_buf;

	control_message             = CMSG_FIRSTHDR(&message);
	control_message->cmsg_level = SOL_SOCKET;
	control_message->cmsg_type  = SCM_RIGHTS;
	control_message->cmsg_len   = CMSG_LEN(sizeof(int));

	memcpy(CMSG_DATA(control_message), &fd_to_send, sizeof(fd_to_send));

	return sendmsg(socket, &message, 0);
}

static tlssep_status_t
_tlssep_connect_or_accept (tlssep_op_t op,
                           tlssep_desc_t *desc,
                           int file_descriptor,
                           const char *expected_common_name,
                           char *common_name)
{
	assert(TLSSEP_OP_CONNECT == op || TLSSEP_OP_ACCEPT == op);
	assert(NULL != desc);
	assert(file_descriptor >= 0);
	assert(NULL != expected_common_name || TLSSEP_OP_ACCEPT == op);

	static tlssep_args_t args; /* static: initialize to zero once to satisfy valgrind [1]. */
	int sockpair[] = { -1, -1 };
	tlssep_status_t status = TLSSEP_STATUS_OK;

	args.op = op;
	if (NULL != expected_common_name) {
		assert(MAXHOSTNAMELEN < sizeof(args.buf));
		strncpy(args.buf, expected_common_name, MAXHOSTNAMELEN);
	} else {
		args.buf[0] = 0x00;
	}
	args.size = NULL == common_name ? false : true;
	int ret = write(desc->context->opSocket, &args, sizeof(args));
	if (sizeof(args) != ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	if (-1 == sendFileDescriptor(desc->context->fdTransferSocket, file_descriptor)) {
		status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	close(file_descriptor);

	if (-1 == socketpair(AF_UNIX, SOCK_STREAM, 0, sockpair)) {
		status = TLSSEP_STATUS_ERROR_SOCKET_FAILED;
		goto done;
	}

	if (-1 == sendFileDescriptor(desc->context->fdTransferSocket, sockpair[1])) {
		status = TLSSEP_STATUS_ERROR_TRANSMITTING_FILE_DESCRIPTOR;
		goto done;
	}

	desc->notificationSocket = sockpair[0];
	if (-1 == fcntl(desc->notificationSocket, F_SETFL, O_NONBLOCK)) {
		status = TLSSEP_STATUS_ERROR_SOCKET_FAILED;
		goto done;
	}

	tlssep_ret_t connect_retval;
	ret = read(desc->context->opSocket, &connect_retval, sizeof(tlssep_ret_t));
	if (sizeof(tlssep_ret_t) != ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	assert(common_name != NULL || args.size == false);

	if (NULL != common_name) {
		if (0 != connect_retval.size) {
			strcpy(common_name, connect_retval.buf);
		} else {
			common_name[0] = 0x00;
		}
	}

	status   = connect_retval.status;
	desc->id = connect_retval.id;

done:
	if (TLSSEP_STATUS_OK != status) {
		if (-1 != desc->notificationSocket) {
			close(desc->notificationSocket);
		}
	}

	if (-1 != sockpair[1]) {
		close(sockpair[1]);
	}

	return status;
}

tlssep_status_t
tlssep_connect (tlssep_context_t *context,
                int file_descriptor,
                const char *expected_common_name,
                char *common_name,
                tlssep_desc_t *desc)
{
	assert(NULL != desc);

	desc->context = context;
	return _tlssep_connect_or_accept(TLSSEP_OP_CONNECT, desc, file_descriptor, expected_common_name, common_name);
}

tlssep_status_t
tlssep_accept(tlssep_context_t *context,
              int file_descriptor,
              const char *expected_common_name,
              char *common_name,
              tlssep_desc_t *desc)
{
	assert(NULL != desc);

	desc->context = context;
	return _tlssep_connect_or_accept(TLSSEP_OP_ACCEPT, desc, file_descriptor, expected_common_name, common_name);
}

tlssep_status_t
tlssep_write (tlssep_desc_t *desc, const void *buf, int buf_size, int *bytes_written)
{
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_written);

	static tlssep_args_t args; /* static: initialize to zero once to satisfy valgrind [1]. */
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (0 == buf_size) {
		goto done;
	}

	args = (tlssep_args_t) {
		.op = TLSSEP_OP_WRITE, .id = desc->id, .size = buf_size
	};
	memcpy(args.buf, buf, buf_size > BUFSIZ ? BUFSIZ : buf_size);
	int ret = write(desc->context->opSocket, &args, sizeof(args));
	if (sizeof(args) != ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	if (buf_size > BUFSIZ) {
		/* Original request larger than args.buf. */
		ret = write(desc->context->opSocket, buf + BUFSIZ, buf_size - BUFSIZ);
		if (buf_size - BUFSIZ != ret) {
			status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
			goto done;
		}
	}

	tlssep_ret_t write_retval;
	ret = read(desc->context->opSocket, &write_retval, sizeof(tlssep_ret_t));
	if (-1 == ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	status = write_retval.status;
	*bytes_written = write_retval.size;

done:
	return status;
}

static tlssep_status_t
_read_ready(tlssep_desc_t *desc, bool *ready)
{
	char token;
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (NULL != ready) {
		*ready = false;
	}

	/* Read token(s) from notification file descriptor. This file descriptor does
	 * not block, so it is okay if there is not data.
	 */
	int count = 0;
	do {
		count = read(desc->notificationSocket, &token, 1);
		if (-1 == count && errno != EAGAIN) {
			status = TLSSEP_STATUS_ERROR_POLL_FAILED;
			goto done;
		} else if (count > 0) {
			if (NULL != ready) {
				*ready = true;
			}
		}
	} while (count != 0 && errno != EAGAIN);

done:
	return status;
}

static tlssep_status_t
tlssep_read_op (tlssep_op_t read_op, tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	assert(TLSSEP_OP_READ == read_op || TLSSEP_OP_PEEK == read_op);
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	static tlssep_args_t args; /* static: initialize to zero once to satisfy valgrind [1]. */
	tlssep_status_t status = TLSSEP_STATUS_OK;

	if (TLSSEP_STATUS_OK != _read_ready(desc, NULL)) {
		status = TLSSEP_STATUS_ERROR;
		goto done;
	}

	if (buf_size == 0) {
		goto done;
	}

	args = (tlssep_args_t) {
		.op = read_op, .id = desc->id, .size = buf_size
	};
	int ret = write(desc->context->opSocket, &args, sizeof(args));
	if (sizeof(args) != ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	tlssep_ret_t read_retval;
	ret = read(desc->context->opSocket, &read_retval, sizeof(tlssep_ret_t));
	if (-1 == ret) {
		status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	if (read_retval.size > 0) {
		if (read_retval.size <= BUFSIZ) {
			memcpy(buf, read_retval.buf, read_retval.size);
		} else {
			ret = read(desc->context->opSocket, buf, read_retval.size);
			if (read_retval.size != ret) {
				status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
				goto done;
			}
		}
	}

	status = read_retval.status;
	if(TLSSEP_STATUS_OK != status) {
		goto done;
	}

	*bytes_read = read_retval.size;

done:
	return status;
}

tlssep_status_t
tlssep_read (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	return tlssep_read_op(TLSSEP_OP_READ, desc, buf, buf_size, bytes_read);
}

tlssep_status_t
tlssep_peek (tlssep_desc_t *desc, void *buf, int buf_size, int *bytes_read)
{
	assert(NULL != desc);
	assert(NULL != buf);
	assert(NULL != bytes_read);

	return tlssep_read_op(TLSSEP_OP_PEEK, desc, buf, buf_size, bytes_read);
}

tlssep_status_t
tlssep_poll (tlssep_desc_t *desc, unsigned int timeout)
{
	assert(NULL != desc);
	assert(-1   != desc->notificationSocket);

	tlssep_status_t status = TLSSEP_STATUS_OK;

	bool ready;
	if (TLSSEP_STATUS_OK != _read_ready(desc, &ready)) {
		status = TLSSEP_STATUS_ERROR_POLL_FAILED;
		goto done;
	}

	if (true == ready) {
		goto done;
	}

	fd_set rfds, wfds;
	struct timeval tv;
	int retval;

	FD_ZERO(&rfds);
        FD_SET(desc->notificationSocket, &rfds);

	FD_ZERO(&wfds);
        FD_SET(desc->notificationSocket, &wfds);

	tv.tv_sec  = timeout;
        tv.tv_usec = 0;

	/* FIXME: Change API to support poll read and/or write; see use in wget. */
        retval = select(desc->notificationSocket + 1, &rfds, &wfds, NULL, &tv);

	status = retval != -1 ? TLSSEP_STATUS_OK : TLSSEP_STATUS_ERROR_POLL_FAILED;

done:
	return status;
}

tlssep_status_t
tlssep_setnonblock (tlssep_desc_t *desc)
{
	assert(NULL != desc);

	tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	static tlssep_args_t args = { .op = TLSSEP_OP_SETNONBLOCK };
	args.id = desc->id;

	int ret = write(desc->context->opSocket, &args, sizeof(args));
	if (sizeof(desc->id) != ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	ret = read(desc->context->opSocket, &retval, sizeof(tlssep_ret_t));
	if (sizeof(tlssep_ret_t) != ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

done:
	return retval.status;
}

tlssep_status_t
tlssep_close (tlssep_desc_t *desc)
{
	assert(NULL != desc);

	static tlssep_args_t args = { .op = TLSSEP_OP_CLOSE };
	tlssep_ret_t retval = { .status = TLSSEP_STATUS_OK };

	args.id = desc->id;

	int ret = write(desc->context->opSocket, &args, sizeof(args));
	if (sizeof(args) != ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

	ret = read(desc->context->opSocket, &retval, sizeof(tlssep_ret_t));
	if (-1 == ret) {
		retval.status = TLSSEP_STATUS_ERROR_RPC_FAILURE;
		goto done;
	}

done:
	if (-1 != desc->notificationSocket) {
		close(desc->notificationSocket);
	}

	return retval.status;
}

tlssep_status_t
tlssep_terminate (tlssep_context_t *context)
{
	assert(NULL != context);

	return _tear_down_decorator(context);
}

#include <tlssep-status-private.h>

char *
tlssep_strerror (tlssep_status_t code)
{
	int num_errors = sizeof(error_message) / sizeof(error_message_t) - 1;

	int i;
	for (i = 0; i < num_errors; i++) {
		if (error_message[i].code == code) {
			break;
		}
	}

	return error_message[i].message;
}

/* [1] structs might have padding, and valgrind complains when padding is not initialized;
 *     use static to initialize to zero once and avoid bzero on each call.
 */
